@@ -9,7 +9,11 @@ import h2.events
 import h2.exceptions
 import h2.settings
 
-from .._exceptions import ConnectionNotAvailable, RemoteProtocolError
+from .._exceptions import (
+    ConnectionNotAvailable,
+    LocalProtocolError,
+    RemoteProtocolError,
+)
 from .._models import Origin, Request, Response
 from .._synchronization import AsyncLock, AsyncSemaphore
 from .._trace import Trace
@@ -37,7 +41,10 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
     CONFIG = h2.config.H2Configuration(validate_inbound_headers=False)
 
     def __init__(
-        self, origin: Origin, stream: AsyncNetworkStream, keepalive_expiry: float = None
+        self,
+        origin: Origin,
+        stream: AsyncNetworkStream,
+        keepalive_expiry: typing.Optional[float] = None,
     ):
         self._origin = origin
         self._network_stream = stream
@@ -52,7 +59,11 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
         self._write_lock = AsyncLock()
         self._sent_connection_init = False
         self._used_all_stream_ids = False
+        self._connection_error = False
         self._events: typing.Dict[int, h2.events.Event] = {}
+        self._read_exception: typing.Optional[Exception] = None
+        self._write_exception: typing.Optional[Exception] = None
+        self._connection_error_event: typing.Optional[h2.events.Event] = None
 
     async def handle_async_request(self, request: Request) -> Response:
         if not self.can_handle_request(request.url.origin):
@@ -111,11 +122,28 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
                 content=HTTP2ConnectionByteStream(self, request, stream_id=stream_id),
                 extensions={"stream_id": stream_id, "http_version": b"HTTP/2"},
             )
-        except Exception:  # noqa: PIE786
+        except Exception as exc:  # noqa: PIE786
             kwargs = {"stream_id": stream_id}
             async with Trace("http2.response_closed", request, kwargs):
                 await self._response_closed(stream_id=stream_id)
-            raise
+
+            if isinstance(exc, h2.exceptions.ProtocolError):
+                # One case where h2 can raise a protocol error is when a
+                # closed frame has been seen by the state machine.
+                #
+                # This happens when one stream is reading, and encounters
+                # a GOAWAY event. Other flows of control may then raise
+                # a protocol error at any point they interact with the 'h2_state'.
+                #
+                # In this case we'll have stored the event, and should raise
+                # it as a RemoteProtocolError.
+                if self._connection_error_event:
+                    raise RemoteProtocolError(self._connection_error_event)
+                # If h2 raises a protocol error in some other state then we
+                # must somehow have made a protocol violation.
+                raise LocalProtocolError(exc)  # pragma: nocover
+
+            raise exc
 
     async def _send_connection_init(self, request: Request) -> None:
         """
@@ -145,7 +173,7 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
         ]
 
         self._h2_state.initiate_connection()
-        self._h2_state.increment_flow_control_window(2 ** 24)
+        self._h2_state.increment_flow_control_window(2**24)
         await self._write_outgoing_data(request)
 
     # Sending the request...
@@ -175,7 +203,7 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
         ]
 
         self._h2_state.send_headers(stream_id, headers, end_stream=end_stream)
-        self._h2_state.increment_flow_control_window(2 ** 24, stream_id=stream_id)
+        self._h2_state.increment_flow_control_window(2**24, stream_id=stream_id)
         await self._write_outgoing_data(request)
 
     async def _send_request_body(self, request: Request, stream_id: int) -> None:
@@ -231,19 +259,39 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
         self, request: Request, stream_id: int
     ) -> h2.events.Event:
         while not self._events.get(stream_id):
-            await self._receive_events(request)
-        return self._events[stream_id].pop(0)
+            await self._receive_events(request, stream_id)
+        event = self._events[stream_id].pop(0)
+        # The StreamReset event applies to a single stream.
+        if hasattr(event, "error_code"):
+            raise RemoteProtocolError(event)
+        return event
 
-    async def _receive_events(self, request: Request) -> None:
-        events = await self._read_incoming_data(request)
-        for event in events:
-            event_stream_id = getattr(event, "stream_id", 0)
+    async def _receive_events(
+        self, request: Request, stream_id: typing.Optional[int] = None
+    ) -> None:
+        async with self._read_lock:
+            if self._connection_error_event is not None:  # pragma: nocover
+                raise RemoteProtocolError(self._connection_error_event)
 
-            if hasattr(event, "error_code"):
-                raise RemoteProtocolError(event)
+            # This conditional is a bit icky. We don't want to block reading if we've
+            # actually got an event to return for a given stream. We need to do that
+            # check *within* the atomic read lock. Though it also need to be optional,
+            # because when we call it from `_wait_for_outgoing_flow` we *do* want to
+            # block until we've available flow control, event when we have events
+            # pending for the stream ID we're attempting to send on.
+            if stream_id is None or not self._events.get(stream_id):
+                events = await self._read_incoming_data(request)
+                for event in events:
+                    event_stream_id = getattr(event, "stream_id", 0)
 
-            if event_stream_id in self._events:
-                self._events[event_stream_id].append(event)
+                    # The ConnectionTerminatedEvent applies to the entire connection,
+                    # and should be saved so it can be raised on all streams.
+                    if hasattr(event, "error_code") and event_stream_id == 0:
+                        self._connection_error_event = event
+                        raise RemoteProtocolError(event)
+
+                    if event_stream_id in self._events:
+                        self._events[event_stream_id].append(event)
 
         await self._write_outgoing_data(request)
 
@@ -262,7 +310,7 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
     async def aclose(self) -> None:
         # Note that this method unilaterally closes the connection, and does
         # not have any kind of locking in place around it.
-        # For task-safe/thread-safe operations call into 'attempt_close' instead.
+        self._h2_state.close_connection()
         self._state = HTTPConnectionState.CLOSED
         await self._network_stream.aclose()
 
@@ -274,11 +322,29 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
         timeouts = request.extensions.get("timeout", {})
         timeout = timeouts.get("read", None)
 
-        async with self._read_lock:
+        if self._read_exception is not None:
+            raise self._read_exception  # pragma: nocover
+
+        try:
             data = await self._network_stream.read(self.READ_NUM_BYTES, timeout)
             if data == b"":
                 raise RemoteProtocolError("Server disconnected")
-            return self._h2_state.receive_data(data)
+        except Exception as exc:
+            # If we get a network error we should:
+            #
+            # 1. Save the exception and just raise it immediately on any future reads.
+            #    (For example, this means that a single read timeout or disconnect will
+            #    immediately close all pending streams. Without requiring multiple
+            #    sequential timeouts.)
+            # 2. Mark the connection as errored, so that we don't accept any other
+            #    incoming requests.
+            self._read_exception = exc
+            self._connection_error = True
+            raise exc
+
+        events = self._h2_state.receive_data(data)
+
+        return events
 
     async def _write_outgoing_data(self, request: Request) -> None:
         timeouts = request.extensions.get("timeout", {})
@@ -286,7 +352,24 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
 
         async with self._write_lock:
             data_to_send = self._h2_state.data_to_send()
-            await self._network_stream.write(data_to_send, timeout)
+
+            if self._write_exception is not None:
+                raise self._write_exception  # pragma: nocover
+
+            try:
+                await self._network_stream.write(data_to_send, timeout)
+            except Exception as exc:  # pragma: nocover
+                # If we get a network error we should:
+                #
+                # 1. Save the exception and just raise it immediately on any future write.
+                #    (For example, this means that a single write timeout or disconnect will
+                #    immediately close all pending streams. Without requiring multiple
+                #    sequential timeouts.)
+                # 2. Mark the connection as errored, so that we don't accept any other
+                #    incoming requests.
+                self._write_exception = exc
+                self._connection_error = True
+                raise exc
 
     # Flow control...
 
@@ -315,7 +398,9 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
 
     def is_available(self) -> bool:
         return (
-            self._state != HTTPConnectionState.CLOSED and not self._used_all_stream_ids
+            self._state != HTTPConnectionState.CLOSED
+            and not self._connection_error
+            and not self._used_all_stream_ids
         )
 
     def has_expired(self) -> bool:
@@ -351,9 +436,9 @@ class AsyncHTTP2Connection(AsyncConnectionInterface):
 
     async def __aexit__(
         self,
-        exc_type: typing.Type[BaseException] = None,
-        exc_value: BaseException = None,
-        traceback: types.TracebackType = None,
+        exc_type: typing.Optional[typing.Type[BaseException]] = None,
+        exc_value: typing.Optional[BaseException] = None,
+        traceback: typing.Optional[types.TracebackType] = None,
     ) -> None:
         await self.aclose()
 
@@ -365,16 +450,26 @@ class HTTP2ConnectionByteStream:
         self._connection = connection
         self._request = request
         self._stream_id = stream_id
+        self._closed = False
 
     async def __aiter__(self) -> typing.AsyncIterator[bytes]:
         kwargs = {"request": self._request, "stream_id": self._stream_id}
-        async with Trace("http2.receive_response_body", self._request, kwargs):
-            async for chunk in self._connection._receive_response_body(
-                request=self._request, stream_id=self._stream_id
-            ):
-                yield chunk
+        try:
+            async with Trace("http2.receive_response_body", self._request, kwargs):
+                async for chunk in self._connection._receive_response_body(
+                    request=self._request, stream_id=self._stream_id
+                ):
+                    yield chunk
+        except BaseException as exc:
+            # If we get an exception while streaming the response,
+            # we want to close the response (and possibly the connection)
+            # before raising that exception.
+            await self.aclose()
+            raise exc
 
     async def aclose(self) -> None:
-        kwargs = {"stream_id": self._stream_id}
-        async with Trace("http2.response_closed", self._request, kwargs):
-            await self._connection._response_closed(stream_id=self._stream_id)
+        if not self._closed:
+            self._closed = True
+            kwargs = {"stream_id": self._stream_id}
+            async with Trace("http2.response_closed", self._request, kwargs):
+                await self._connection._response_closed(stream_id=self._stream_id)
